//
// A Simple Mesh Viewer & Line Test Application
//
// By Jonathan Kreuzer
//
// All of the windows specific stuff is in this file, plus the RenderFrame & SetLine functions
// 
//
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <gl/gl.h>
#include <gl/glu.h>
#include "resource.h"
#include <math.h>
#include <string>
#include "vec3.h"
#include "matrix.h"
#include "renderInterface.h"
#include "mesh.h"

// Structs & Globals
struct SView {
	SView() {
		RotateX = 90.0f;
		RotateY = 0.0f;
	}
	float RotateX;
	float RotateY;
};
SView g_View;

CRenderer *g_pRenderer = NULL;
CMesh g_Mesh;
CVec3 g_LineP[2];
int g_WindowHeight;
		
// Functions
ATOM				MyRegisterClass(HINSTANCE hInstance);
LRESULT CALLBACK	WndProc(HWND, UINT, WPARAM, LPARAM);
void RenderFrame( class CRenderer *pRender );

// Main
int WINAPI WinMain(HINSTANCE hInstance,
                     HINSTANCE hPrevInstance,
                     LPTSTR    lpCmdLine,
                     int       nCmdShow)
{
	MyRegisterClass(hInstance);
	
	HWND hWnd = CreateWindow( "testappclass", "Sample App: stage 1 (mesh viewer + selection)", WS_OVERLAPPEDWINDOW,
				 CW_USEDEFAULT, 0, 800, 600, NULL, NULL, hInstance, NULL);

    if (!hWnd) return FALSE; 

	// Initialize Renderer
	CRenderer *pRender = new CRenderer;
	g_pRenderer = pRender;
	pRender->InitRenderWindow( hWnd, true, true );
	if ( !pRender->IsInitialized() ) return 0;

	// Show the Window
	MoveWindow( hWnd, 20, 20, 800, 600, TRUE );
    ShowWindow(hWnd, nCmdShow);
    UpdateWindow(hWnd);
	
	// Load the Mesh	
	g_Mesh.Load( "meshhead.dat" );
	
	// Main loop
	MSG msg;
	bool bDone = false;
	
	while ( !bDone ) 
	{
		// Render the Frame
		RenderFrame( pRender );	
		
		// Check for and process windows messages each frame
		if ( PeekMessage(&msg, NULL, 0, 0, PM_REMOVE ) )
		{		
			if ( msg.message == WM_QUIT ) {
				bDone = true;
			} else {	
				TranslateMessage(&msg);
				DispatchMessage(&msg);
			}
		}
	}

	// Exit
	pRender->Shutdown( );
	return (int) msg.wParam;
}
//
// 
//
void RenderFrame( CRenderer *pRender )
	{
	// Rotate the Viewport
	CMatrix ModelView;
	ModelView.Translate( CVec3(0, 0, -800.0f) );
	ModelView.Rotate( -g_View.RotateY, 1, 0, 0 );
	ModelView.Rotate( -g_View.RotateX, 0, 1, 0 );		
	
	// Render
	pRender->ClearBuffer( );	
	pRender->SetMatrix( ModelView );		
	pRender->SetColor( 0.0f, .75f, 0.0f );
	pRender->DrawLine( g_LineP[0], g_LineP[1] );
	pRender->SetColor( 0.0f, .4f, .8f );
	g_Mesh.Render( pRender );
	g_Mesh.CheckLine( g_LineP[0], g_LineP[1] );		

	// Show the back buffer
	pRender->BufferSwap( );		
	}
//
// Menu command processing
// There will be more here later.
//
int MenuCommand( HWND hWnd, int wmId, int wmEvent )
{
switch (wmId)
	{
	case IDM_EXIT:
		DestroyWindow(hWnd);
	return TRUE;
	}
return FALSE;
}

//
// This function will find 2 points in world space that are on the line into the screen defined by windowspace point (x,y)
//
void SetLine( int x, int y )
{
   double mvmatrix[16];
   double projmatrix[16];
   int viewport[4];
   double dX, dY, dZ, dClickY; // glUnProject uses doubles, but I'm using floats for these 3D vectors
	
   glGetIntegerv(GL_VIEWPORT, viewport);	
   glGetDoublev (GL_MODELVIEW_MATRIX, mvmatrix);
   glGetDoublev (GL_PROJECTION_MATRIX, projmatrix);
   dClickY = double (g_WindowHeight -y); // fixme
   // OpenGL renders with (0,0) on bottom, mouse reports with (0,0) on top
	
   gluUnProject ((double) x, dClickY, 0.0, mvmatrix, projmatrix, viewport, &dX, &dY, &dZ);
   g_LineP[0] = CVec3 ( (float) dX, (float) dY, (float) dZ );
   gluUnProject ((double) x, dClickY, 1.0, mvmatrix, projmatrix, viewport, &dX, &dY, &dZ);
   g_LineP[1] = CVec3 ( (float) dX, (float) dY, (float) dZ );
}
//
// 
//
LRESULT CALLBACK WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	PAINTSTRUCT ps;
	HDC hdc;
	POINT Point;
	static bool bRight = false;
	static POINT StartPoint = {0,0};
	
	switch (message)
	{
	case WM_COMMAND:
		if ( !MenuCommand( hWnd, LOWORD(wParam), HIWORD(wParam) ) )
			return DefWindowProc(hWnd, message, wParam, lParam);
		break;
		
	case WM_ERASEBKGND:
		// This window is fully drawn by OpenGL, so don't allow any standard window erasing
		return false;	
			
	case WM_PAINT:
		hdc = BeginPaint(hWnd, &ps);
		if (g_pRenderer && g_pRenderer->IsInitialized() ) 
			{
			RenderFrame( g_pRenderer );
			}
		EndPaint(hWnd, &ps);
		break;
		
	case WM_DESTROY:
		PostQuitMessage(0);
		break;

	case WM_LBUTTONDOWN: 
		// Send mouse location in window coordinates
        SetLine( LOWORD(lParam), HIWORD(lParam) );
		break;
		
	case WM_LBUTTONUP:
		break;
				
	case WM_RBUTTONDOWN:
	
		GetCursorPos(&StartPoint); // mouse location for the full screen coordinates
		bRight = true;	
		break;
		
	case WM_RBUTTONUP:
		bRight = false;
		break;
		
	case WM_SIZE:

		g_pRenderer->SetSize( LOWORD(lParam), HIWORD(lParam) );
		g_WindowHeight = HIWORD(lParam); // keep track of the window height so we can flip the mouse y	
		break;
	
	case WM_MOUSEMOVE:
		if (bRight) {
			GetCursorPos(&Point);
		
			g_View.RotateX += (Point.x - StartPoint.x) / 2.0f;
			g_View.RotateY += (Point.y - StartPoint.y) / 2.0f;
			StartPoint = Point;
			
			if ( g_View.RotateX > 360 ) g_View.RotateX -=360;
			if ( g_View.RotateX < 0 ) g_View.RotateX +=360;
			if ( g_View.RotateY > 360 ) g_View.RotateY -=360;
			if ( g_View.RotateY < 0 ) g_View.RotateY +=360;
			}
		break;
				
	default:
		return DefWindowProc(hWnd, message, wParam, lParam);
	}
	
	return 0;
}
//
//  
//
ATOM MyRegisterClass(HINSTANCE hInstance)
{
	WNDCLASSEX wcex;
	
	wcex.cbSize = sizeof(WNDCLASSEX);
	wcex.style			= CS_HREDRAW | CS_VREDRAW;
	wcex.lpfnWndProc	= WndProc;
	wcex.cbClsExtra		= 0;
	wcex.cbWndExtra		= 0;
	wcex.hInstance		= hInstance;
	wcex.hIcon			= LoadIcon(hInstance, MAKEINTRESOURCE(IDI_LINEMESHTEST));
	wcex.hCursor		= LoadCursor(NULL, IDC_ARROW);
	wcex.hbrBackground	= (HBRUSH)(COLOR_WINDOW+1);
	wcex.lpszMenuName	= MAKEINTRESOURCE(IDC_LINEMESHTEST);
	wcex.lpszClassName	= "testappclass";
	wcex.hIconSm		= 0;

	return RegisterClassEx(&wcex);
}	